<?php
# @Author: Aaron Bowie
# @Date:   Saturday, September 16th 2017, 12:09:42 am
# @Email:  support@weareag.co.uk
# @Filename: ag-woocommerce-sagepay-class.php
# @Last modified by:   Aaron Bowie
# @Last modified time: Thursday, March 22nd 2018, 10:16:13 pm


defined('ABSPATH') or die("No script kiddies please!");


/*-----------------------------------------------------------------------------------*/
/*  Set class
/*-----------------------------------------------------------------------------------*/

class ag_sagepay_server extends WC_Payment_Gateway {

function __construct() {

	$this->id = "ag_sagepay_server";
	$this->method_title = __( "AG SagePay Server", 'ag-sagepay-server' );
  $license = get_option('AGD_SagePay_Server_Key');
  $websiteurl = get_option ('siteurl');
  $api_params = array(
    'slm_action' => 'slm_check',
    'secret_key' => AGD_KEY,
    'license_key' => $license,
  );
  $query = esc_url_raw(add_query_arg($api_params, AGD));
  $response = wp_remote_get($query, array('timeout' => 20, 'sslverify' => false));
  $license_data = json_decode(wp_remote_retrieve_body($response));

  if($license_data->result == 'success'){
    $this->method_description =  __( "SagePay Server payment gateway plugin for WooCommerce", 'ag-sagepay-server' );
  } else {
    $this->method_description = __( 'Please enter the license key for this product to activate it. You were given a license key when you purchased this item. <br /><strong>You can do this on the AG Plugins menu item on the left</strong>.', 'ag-sagepay-server' );
  }

	$this->title = __( "SagePay Server", 'ag-sagepay-server' );
	$this->icon = plugins_url('AG-Woo-SagePay-Server/img/sagepay-logo.png');
	$this->has_fields = false;
	$this->notify_url   = str_replace( 'https:', 'http:', home_url( '/wc-api/woocommerce_ag_sagepay_server' ) );

	// Supports
	$this->supports = array( 'refunds' );

  // actions
	add_action( 'woocommerce_api_woocommerce_ag_sagepay_server', array( $this, 'successful_request' ) );
	add_action( 'woocommerce_receipt_ag_sagepay_server', array( $this, 'receipt_page' ) );

  if($license_data->result == 'success'){
    $this->init_form_fields();
    $this->init_settings();
  }

	foreach ( $this->settings as $setting_key => $value ) {
		$this->$setting_key = $value;
	}

	$this->environment 		= $this->get_option('environment');

	if ( $this->environment == "yes" ) {
		$this->description 	= $this->get_option( 'description' );
		$this->description .= '<br /><br />' . sprintf( __( '<br /><br /><strong>TEST MODE ACTIVE.</strong><br />In test mode you can run test payments using card type VISA, card number <strong>4929 0000 0000 6</strong> with any CVC number and a valid expiration date. Check SagePay docs for more cards to try .<br /><br />If 3DSecure is enabled on the account you can test this also, the password for testing is password. <a target="_blank" href="%s">here</a> for more card numbers & steps on setting up.' ), 'https://www.weareag.co.uk/product/sage-pay-server-woocommerce/' );
		$this->description  = trim( $this->description );
	} else {
		$this->description 	= $this->get_option( 'description' );
	}

	// Save settings
	if ( is_admin() ) {
		add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );
	}
} // End __construct()

/*-----------------------------------------------------------------------------------*/
/*  Build setting fields
/*-----------------------------------------------------------------------------------*/

public function init_form_fields() {

  $this->sage_cardtypes = apply_filters( 'ag_sagepaydirect_cardtypes', array(
    'MasterCard'		    => __( 'MasterCard Credit', 'ag-sagepay-server' ),
    'MasterCard Debit'	=> __( 'MasterCard Debit', 'ag-sagepay-server' ),
    'Visa'				      => __( 'Visa Credit', 'ag-sagepay-server' ),
    'Visa Debit'		    => __( 'Visa Debit', 'ag-sagepay-server' ),
    'Discover'			    => __( 'Discover', 'ag-sagepay-server' ),
    'American Express' 	=> __( 'American Express', 'ag-sagepay-server' ),
    'Maestro'			      => __( 'Maestro', 'ag-sagepay-server' ),
    'JCB'				        => __( 'JCB', 'ag-sagepay-server' ),
    'Laser'				      => __( 'Laser', 'ag-sagepay-server' ),
    //'PayPal'				  => __( 'PayPal', 'ag-sagepay-server' ),
  ) );

  $this->form_fields = array(
		'enabled'   => array(
			'title'		=> __( 'Enable / Disable', 'ag-sagepay-server' ),
			'label'		=> __( 'Enable this payment gateway', 'ag-sagepay-server' ),
			'type'		=> 'checkbox',
			'default'	=> 'no',
		),
		'title' => array(
			'title'		=> __( 'Title', 'ag-sagepay-server' ),
			'type'		=> 'text',
			'desc_tip'	=> __( 'Payment title the customer will see during the checkout process.', 'ag-sagepay-server' ),
			'default'	=> __( 'Credit card', 'ag-sagepay-server' ),
		),
		'description' => array(
			'title'		  => __( 'Description', 'ag-sagepay-server' ),
			'type'		  => 'textarea',
			'desc_tip'	=> __( 'Payment description the customer will see during the checkout process.', 'ag-sagepay-server' ),
			'default'	  => __( 'Pay securely using your credit card.', 'ag-sagepay-server' ),
			'css'		    => 'max-width:350px;'
		),
    'vendor' => array(
      'title'		  => __( 'SagePay Server Vendor Name', 'ag-sagepay-server' ),
      'type'		  => 'text',
      'desc_tip'	=> __( 'This is the Vendor Name provided by SagePay Server when you signed up for an account.', 'ag-sagepay-server' ),
    ),
		'vendorpwd'         => array(
        'title'         => __( 'Encryption Password', 'ag-sagepay-server' ),
        'type'          => 'text',
        'description'   => __( 'This cane be found in your SagePay account. This NOT the vendor password', 'ag-sagepay-server' ),
      'desc_tip'    	=> true,
    ),
    'cardtypes'	=> array(
      'title' 		=> __( 'Accepted Cards', 'ag-sagepay-server' ),
      'type' 			=> 'multiselect',
      'class'			=> 'chosen_select',
      'css'         	=> 'width: 350px;',
      'description' 	=> __( 'Select which card types to accept.', 'ag-sagepay-server' ),
      'default' 		=> '',
      'options' 		=> $this->sage_cardtypes,
    ),
    'transtype' => array(
      'title' => __('Transition Type', 'ag-sagepay-server'),
      'type' => 'select',
      'css'  => 'height: 35px;',
      'options' => array(
        'PAYMENT' => __('Payment', 'ag-sagepay-server'),
        //'DEFFERRED' => __('Deferred', 'ag-sagepay-server'),
        'AUTHENTICATE' => __('Authenticate', 'ag-sagepay-server')
      ),
      'description' => __('Select Payment or Authenticated.', 'ag-sagepay-server')
    ),
		'environment' => array(
			'title'		=> __( 'SagePay Server Test Mode', 'ag-sagepay-server' ),
			'label'		=> __( 'Enable Test Mode', 'ag-sagepay-server' ),
			'type'		=> 'checkbox',
			'description' => __( 'Place the payment gateway in test mode.', 'ag-sagepay-server' ),
			'default'	=> 'no',
		),
		'password' => array(
			'title'		=> __( 'SagePay Password', 'ag-sagepay-server' ),
			'type'		=> 'password',
			'description' => __( 'This is used for processing refunds only, <strong>NOT</strong> needed to process payments.', 'ag-sagepay-server' ),
		),
		'user' => array(
			'title'		=> __( 'SagePay User', 'ag-sagepay-server' ),
			'type'		=> 'text',
			'description' => __( 'This is used for processing refunds only, <strong>NOT</strong> needed to process payments.', 'ag-sagepay-server' ),
		),
	);
}




public function generate_sagepayform_form( $order_id ) {

	global $woocommerce;

		$order = new WC_Order( $order_id );


		$gateway_url = ( "FALSE" == $environment )
		 ? 'https://live.sagepay.com/gateway/service/vspform-register.vsp'
		 : 'https://test.sagepay.com/gateway/service/vspform-register.vsp';

			$basket = '';
			$item_loop = 0;
			if ( sizeof( $order->get_items() ) > 0 ) {
				foreach ( $order->get_items() as $item ) {
					if ( $item['qty'] ) {

						$item_loop++;

						$product = $order->get_product_from_item( $item );

						$item_name 	= $item['name'];

						$item_meta = new WC_Order_Item_Meta( $item['item_meta'] );
						if ( $meta = $item_meta->display( true, true ) )
							$item_name .= ' ( ' . $meta . ' )';

						$item_cost = $order->get_item_subtotal( $item, false );
						$item_total_inc_tax = $order->get_item_subtotal( $item, true )*$item['qty'];
						$item_total = $order->get_item_subtotal( $item, false )*$item['qty'];
						$item_tax = number_format( (float) ($item_total_inc_tax - $item_total)/$item['qty'], 2, '.', '' );

						if($item_loop > 1){
							$basket .= ':';
						}
						$sku = '';
						if ( $product->get_sku() ) {
							$sku = '['.$product->get_sku().']';
						}

						$basket .= str_replace(':',' = ',$sku).str_replace(':',' = ',$item_name).':'.$item['qty'].':'.$item_cost.':'.$item_tax.':'.number_format( $item_cost+$item_tax, 2, '.', '' ).':'.$item_total_inc_tax;
					}
				}
			}


			$item_loop++;

			$basket .= ':Order Total:---:---:---:---:'.$order->get_total();

			$basket = $item_loop.':'.$basket;

			$time_stamp = date("ymdHis");
			$orderid = $this->vendor . "-" . $time_stamp . "-" . $order_id;
			WC()->session->set('sagepay_vtc', $orderid);

			$ag_sps_arg['ReferrerID'] 					= '495E035E-B1B8-4BF7-82F0-5DEE0D6A380C';
			$ag_sps_arg['Amount'] 							= $order->get_total();
			$ag_sps_arg['CustomerName']					= $customer_order->billing_first_name . ' ' . $customer_order->billing_last_name;
			$ag_sps_arg['CustomerEMail'] 				= $order->billing_email;
			$ag_sps_arg['BillingFirstnames'] 		= $order->billing_first_name;
			$ag_sps_arg['BillingSurname'] 			= $order->billing_last_name;
			$ag_sps_arg['BillingAddress1'] 			= $order->billing_address_1;
			$ag_sps_arg['BillingAddress2'] 			= $order->billing_address_2;
			$ag_sps_arg['BillingCity'] 					= $order->billing_city;

			if( $order->billing_country == 'US' ){
				$ag_sps_arg['BillingState'] 			= $order->billing_state;
			}else{
				$ag_sps_arg['BillingState'] 			= '';
			}

			$ag_sps_arg['BillingPostCode'] 			= $order->billing_postcode;
			$ag_sps_arg['BillingCountry'] 			= $order->billing_country;
			$ag_sps_arg['BillingPhone'] 				= $order->billing_phone;

			$ag_sps_arg['DeliveryFirstnames'] 	= $order->shipping_first_name;
			$ag_sps_arg['DeliverySurname'] 			= $order->shipping_last_name;
			$ag_sps_arg['DeliveryAddress1'] 		= $order->shipping_address_1;
			$ag_sps_arg['DeliveryAddress2'] 		= $order->shipping_address_2;
			$ag_sps_arg['DeliveryCity'] 				= $order->shipping_city;

			if( $order->shipping_country == 'US' ){
				$ag_sps_arg['DeliveryState'] 			= $order->shipping_state;
			}else{
				$ag_sps_arg['DeliveryState'] 			= '';
			}

			$ag_sps_arg['DeliveryPostCode'] 		= $order->shipping_postcode;
			$ag_sps_arg['DeliveryCountry'] 			= $order->shipping_country;


			$ag_sps_arg['DeliveryPhone'] 					= $order->shipping_phone;
			$ag_sps_arg['FailureURL'] 						= $this->notify_url;
			$ag_sps_arg['SuccessURL'] 						= $this->notify_url;
			$ag_sps_arg['Description'] 						= sprintf(__('Order #%s' , 'ag-sagepay-server'), ltrim( $order->get_order_number(), '#' ));
			$ag_sps_arg['Currency'] 							= get_woocommerce_currency();
			$ag_sps_arg['VendorTxCode'] 					= $orderid;
			//$ag_sps_arg['VendorEMail'] 		=
			//$ag_sps_arg['SendEMail'] 			=

			$ag_sps_arg['Basket'] 								= $basket;

			$post_values = "";
			foreach( $ag_sps_arg as $key => $value ) {
					$post_values .= "$key=" . trim( $value ) . "&";
			}
			$post_values = substr($post_values, 0, -1);

			$params['VPSProtocol'] = "3.00";
			$params['TxType'] = $this->transtype;
			$params['Vendor'] = $this->vendor;
			$params['Crypt'] = $this->encrypt_data($post_values);

			$ag_sps_arg_array = array();

			foreach ($params as $key => $value) {
				$ag_sps_arg_array[] = '<input type="hidden" name="'.esc_attr( $key ).'" value="'.esc_attr( $value ).'" />';
			}

			wc_enqueue_js('
				jQuery("body").block({
						message: "'.__('You are now redirecting to SagePays server to make payment securly.', 'ag-sagepay-server').'",
						overlayCSS:
						{
							background: "#fff",
							opacity: 0.6
						},
						css: {
									padding:        20,
									textAlign:      "center",
									color:          "#555",
									border:         "3px solid #aaa",
									backgroundColor:"#fff",
									cursor:         "wait",
									lineHeight:		"32px"
							}
					});
				jQuery("#submit_sagepay_payment_form").click();
			');

			return  '<form action="'.esc_url( $gateway_url ).'" method="post" id="sagepay_payment_form">
			' . implode('', $ag_sps_arg_array) . '
			<input type="submit" class="button" id="submit_sagepay_payment_form" value="'.__('Pay via SagePay', 'ag-sagepay-server').'" /> <a class="button cancel" href="'.esc_url( $order->get_cancel_order_url() ).'">'.__('Cancel order &amp; restore cart', 'ag-sagepay-server').'</a>
		</form>';

}

/*-----------------------------------------------------------------------------------*/
/*	Process the payment and return the result
/*-----------------------------------------------------------------------------------*/
function process_payment( $order_id ) {

	$order = new WC_Order( $order_id );

	return array(
		'result' 	=> 'success',
		'redirect'	=> $order->get_checkout_payment_url( true )
	);

}

/*-----------------------------------------------------------------------------------*/
/*	receipt_page
/*-----------------------------------------------------------------------------------*/
function receipt_page( $order ) {

	echo '<p>'.__('Thank you for your order.', 'ag-sagepay-server').'</p>';

	echo $this->generate_sagepayform_form( $order );

}

/*-----------------------------------------------------------------------------------*/
/*	Successful Payment WOOP WOOP!
/*-----------------------------------------------------------------------------------*/
function successful_request() {
	global $woocommerce;

	if ( isset($_REQUEST['crypt']) && !empty($_REQUEST['crypt']) ) {

		$transaction_response = $this->decode(str_replace(' ', '+',$_REQUEST['crypt']));

		$order_id = explode('-',$transaction_response['VendorTxCode']);

		if ( $transaction_response['Status'] == 'OK' || $transaction_response['Status'] == 'AUTHENTICATED'|| $transaction_response['Status'] == 'REGISTERED' ) {

			$order = new WC_Order( $order_id[2] );

			if (isset($transaction_response['Status'])) {
					update_post_meta($order->id, 'Status', $transaction_response['Status']);
			}
			if (isset($transaction_response['StatusDetail'])) {
					update_post_meta($order->id, 'StatusDetail', $transaction_response['StatusDetail']);
			}
			if (isset($transaction_response['VPSTxId'])) {
					update_post_meta($order->id, 'VPSTxId', $transaction_response['VPSTxId']);
			}
			if (isset($transaction_response['CAVV'])) {
					update_post_meta($order->id, 'CAVV', $transaction_response['CAVV']);
			}
			if (isset($transaction_response['SecurityKey'])) {
					update_post_meta($order->id, 'SecurityKey', $transaction_response['SecurityKey']);
			}
			if (isset($transaction_response['TxAuthNo'])) {
					update_post_meta($order->id, 'TxAuthNo', $transaction_response['TxAuthNo']);
			}
			if (isset($transaction_response['AVSCV2'])) {
					update_post_meta($order->id, 'AVSCV2', $transaction_response['AVSCV2']);
			}
			if (isset($transaction_response['AddressResult'])) {
					update_post_meta($order->id, 'AddressResult', $transaction_response['AddressResult']);
			}
			if (isset($transaction_response['PostCodeResult'])) {
					update_post_meta($order->id, 'PostCodeResult', $transaction_response['PostCodeResult']);
			}
			if (isset($transaction_response['CV2Result'])) {
					update_post_meta($order->id, 'CV2Result', $transaction_response['CV2Result']);
			}
			if (isset($transaction_response['3DSecureStatus'])) {
					update_post_meta($order->id, '3DSecureStatus', $transaction_response['3DSecureStatus']);
			}
			if (WC()->session->get('sagepay_vtc') != '') {
					update_post_meta($order->id, 'VendorTxCode', WC()->session->get('sagepay_vtc'));
			}

			$order->add_order_note(sprintf(__('SagePay Payment Completed. The Order Reference Number is %s.', 'ag-sagepay-server'), $transaction_response['VPSTxId']));
			$order->payment_complete();
			wp_redirect( $this->get_return_url( $order ) ); exit;

		}else{

			wc_add_notice( sprintf(__('The Transaction Failed. The Error Message was %s', 'ag-sagepay-server'), $transaction_response['StatusDetail'] ), $notice_type = 'error' );
			wp_redirect( get_permalink(get_option( 'woocommerce_checkout_page_id' )) ); exit;

		}
	}
}

/*-----------------------------------------------------------------------------------*/
/*	Refunds
/*-----------------------------------------------------------------------------------*/

public function process_refund( $order_id, $amount = NULL, $reason = '' ) {

    	$order 			   = new WC_Order( $order_id );

			$VendorTxCode  = 'Refund-' . $order_id . '-' . time();
			$ref = get_post_meta( $order->id, 'VPSTxId', true );

			if ( ! $ref ) {
				return new WP_Error( 'error', __( 'Refund failed: Transaction ID not found.', 'ag_epdq_checkout' ) );
			}
			if ( ! $amount ) {
				return new WP_Error( 'error', __( 'Refund failed: Amount invalid.', 'ag_epdq_checkout' ) );
			}

			/*-----------------------------------------------------------------------------------*/
			/*	Getting securitykey form SagePay for refund
			/*-----------------------------------------------------------------------------------*/

			$command = 'getTransactionDetail';
			$vendor = $this->vendor;
			$user = $this->user;
			$password = $this->password;
			$string = '<command>'.$command.'</command><vendor>'.$vendor.'</vendor><user>'.$user.'</user><vpstxid>'. get_post_meta( $order->id, 'VPSTxId', true ). '</vpstxid>';
			$crypt = MD5($string . '<password>' . $password . '</password>');
			$xml = 'XML=<vspaccess>'.$string.'<signature>'.$crypt.'</signature></vspaccess>';
			$access_url = ( "FALSE" == $environment )
								 ? 'https://live.sagepay.com/access/access.htm'
								 : 'https://test.sagepay.com/access/access.htm';
			$results = wp_remote_post($access_url, array(
				'method' => 'POST',
				'timeout'     => 6,
				'redirection' => 5,
				'body' => $xml,
			));
			$body     = wp_remote_retrieve_body($results);
			$xml  = simplexml_load_string($body);
			$securitykey = [ (string) $xml->securitykey ][0];

			// Storing
			if (isset($securitykey)) {
					update_post_meta($order->id, 'SecurityKey', $securitykey);
			}

			/*-----------------------------------------------------------------------------------*/
			/*	END
			/*-----------------------------------------------------------------------------------*/


			// API Request for refund
			$api_request 	 = 'VPSProtocol=' . '3.00';
			$api_request 	.= '&TxType=REFUND';
			$api_request  .= '&Vendor=' . $this->vendor;
			$api_request 	.= '&VendorTxCode=' . $VendorTxCode;
			$api_request  .= '&Amount=' . urlencode( $amount );
			$api_request 	.= '&Currency=' . get_woocommerce_currency();
			$api_request 	.= '&Description=Refund for order ' . $order->id;
			$api_request	.= '&RelatedVPSTxId=' . get_post_meta( $order->id, 'VPSTxId', true );
			$api_request	.= '&RelatedVendorTxCode=' . get_post_meta( $order->id, 'VendorTxCode', true );
			$api_request	.= '&RelatedSecurityKey=' . $securitykey;
			$api_request	.= '&RelatedTxAuthNo=' . get_post_meta( $order->id, 'TxAuthNo', true );

      $refund_url = ( "FALSE" == $environment )
        ? 'https://live.sagepay.com/gateway/service/refund.vsp'
        : 'https://test.sagepay.com/gateway/service/refund.vsp';

			$result = wp_remote_post($refund_url, array(
        'method' => 'POST',
        'timeout'     => 6,
        'redirection' => 5,
        'body' => $api_request,
      ));

			$sagepay_response = array();
      $lines = preg_split('/\r\n|\r|\n/', $result['body']);
      foreach ($lines as $line) {
        $key_value = preg_split('/=/', $line, 2);
        if (count($key_value) > 1) {
          $sagepay_response[trim($key_value[0])] = trim($key_value[1]);
        }
      }

			if ( $sagepay_response['Status'] == "OK" ) {
        $order->add_order_note( __('Refund successful', 'ag-sagepay-direct') . '<br />' .
          __('Refund Amount : ', 'ag-sagepay-direct') . $amount . '<br />' .
          __('Refund Reason : ', 'ag-sagepay-direct') . $reason . '<br />'
      );
      return true;
			} else {
				$order->add_order_note( __('Refund failed', 'ag-sagepay-direct') . '<br />'  );
			}

}

	private function encrypt_data($strIn) {
		$strIn = $this->pkcs5($strIn, 16);
		return "@".bin2hex(mcrypt_encrypt(MCRYPT_RIJNDAEL_128, $this->vendorpwd, $strIn, MCRYPT_MODE_CBC, $this->vendorpwd));
	}

	private function decrypt_data($strIn) {
		$strIn = substr($strIn, 1);
		$strIn = pack('H*', $strIn);
		return mcrypt_decrypt(MCRYPT_RIJNDAEL_128, $this->vendorpwd, $strIn, MCRYPT_MODE_CBC, $this->vendorpwd);
	}

	private function pkcs5($text, $blocksize)	{
		$pad = $blocksize - (strlen($text) % $blocksize);
		return $text . str_repeat(chr($pad), $pad);
	}

	public function decode($strIn) {
		$decodedString = $this->decrypt_data($strIn);
		parse_str($decodedString, $sagePayResponse);
		return $sagePayResponse;
	}

}
